@@ -7,7 +7,8 @@ import concurrent.futures
 from sqlalchemy import create_engine
 import traceback
 from colorama import Fore, Style
-from os import linesep
+from os import linesep, path
+from jinja2 import Template
 
 
 class DataCheckException(Exception):
@@ -49,6 +50,12 @@ class DataCheck:
         self.workers = workers
         self.verbose = verbose
         self.traceback = traceback
+        self.template_data = {}
+
+    def load_template(self):
+        template_yaml = Path("checks") / "template.yml"
+        if template_yaml.exists():
+            self.template_data = yaml.safe_load(template_yaml.open())
 
     @property
     def executor(self):
@@ -58,17 +65,15 @@ class DataCheck:
         """
         Return parameter specific to a database.
         """
-        if "cx_oracle://" in self.connection or "oracle://" in self.connection:
-            # Fix to remove SAWarning. Should be removed with SQLAlchemy 1.4
-            return {"max_identifier_length": 128}
-        else:
-            return {}
+        return {}  # no special parameters needed for now
 
-    def get_engine(self):
+    def get_engine(self, extra_params={}):
         """
         Return the database engine for the connection.
         """
-        return create_engine(self.connection, **self.get_db_params())
+        return create_engine(
+            path.expandvars(self.connection), **{**self.get_db_params(), **extra_params}
+        )
 
     def run_query(self, query: str) -> pd.DataFrame:
         """
@@ -76,7 +81,8 @@ class DataCheck:
         """
         if not self.connection:
             raise DataCheckException(f"undefined connection: {self.connection}")
-        return pd.read_sql(query, self.get_engine())
+        with self.get_engine().connect() as connection:
+            return pd.read_sql_query(query, connection)
 
     def get_expect_file(self, sql_file: Path) -> Path:
         """
@@ -84,7 +90,7 @@ class DataCheck:
         """
         return sql_file.parent / (sql_file.stem + ".csv")
 
-    def print_failed(self, df: pd.DataFrame) -> str:
+    def print_failed(self, df: pd.DataFrame, print_format="pandas") -> str:
         """
         Prints a DataFrame with diff information and returns it as a string.
         """
@@ -92,7 +98,12 @@ class DataCheck:
             df["_diff"] = ""
             df.loc[df._merge == "left_only", ["_diff"]] = "db"
             df.loc[df._merge == "right_only", ["_diff"]] = "expected"
-            return str(df.drop(["_merge"], axis=1))
+            if print_format == "pandas":
+                return str(df.drop(["_merge"], axis=1))
+            elif print_format.lower() == "csv":
+                return df.drop(["_merge"], axis=1).to_csv(index=False)
+            else:
+                raise DataCheckException(f"unknown print format: {print_format}")
 
     def merge_results(
         self, sql_result: DataFrame, expect_result: DataFrame
@@ -125,8 +136,21 @@ class DataCheck:
     def str_fail(string):
         return Fore.RED + string + Style.RESET_ALL
 
+    def read_sql_file(self, sql_file) -> str:
+        """
+        Reads the SQL file and returns it as a string.
+        Evaluates the templates when needed.
+        """
+        return Template(sql_file.read_text(encoding="UTF-8")).render(
+            **self.template_data
+        )
+
     def run_test(
-        self, sql_file: Path, return_all=False, print_failed=False
+        self,
+        sql_file: Path,
+        return_all=False,
+        print_failed=False,
+        print_format="pandas",
     ) -> DataCheckResult:
         """
         Run a data_check test on a single input file.
@@ -146,7 +170,7 @@ class DataCheck:
             )  # no need to run queries, if no expected results found
 
         try:
-            sql_result = self.run_query(sql_file.read_text(encoding="UTF-8"))
+            sql_result = self.run_query(self.read_sql_file(sql_file))
         except Exception as exc:
             fail = self.str_fail(f"FAILED (with exception in {sql_file})")
             message = f"{sql_file}: {fail}"
@@ -194,7 +218,7 @@ class DataCheck:
             failed = self.str_fail("FAILED")
             message = f"{sql_file}: {failed}"
             if print_failed:
-                message += linesep + self.print_failed(df_diff.copy())
+                message += linesep + self.print_failed(df_diff.copy(), print_format)
             passed = False
         else:
             passed_msg = self.str_pass("PASSED")
@@ -221,21 +245,16 @@ class DataCheck:
                 raise Exception(f"unexpected path: {f}")
         return result
 
-    def run(self, files: List[Path], print_failed=False) -> bool:
+    def run(self, files: List[Path], print_failed=False, print_format="pandas") -> bool:
         """
-        Runs a data_check test for all element in the list in parallel.
+        Runs a data_check test for all element in the list.
         Returns True, if all calls returned True, otherweise False.
         """
         all_files = self.expand_files(files)
-        result_futures = [
-            self.executor.submit(self.run_test, f, print_failed=print_failed)
-            for f in all_files
-        ]
-        results = []
-        for future in concurrent.futures.as_completed(result_futures):
-            dc_result = future.result()
-            results.append(dc_result)
-            print(dc_result.message)
+        if self.workers == 1 or len(all_files) == 1:
+            results = self.run_serial(all_files, print_failed, print_format)
+        else:
+            results = self.run_parallel(all_files, print_failed, print_format)
 
         overall_result = all(results)
         overall_result_msg = (
@@ -245,34 +264,64 @@ class DataCheck:
         print(f"overall result: {overall_result_msg}")
         return overall_result
 
-    def gen_expectation(self, sql_file: Path):
+    def run_parallel(self, all_files, print_failed, print_format):
+        """
+        Runs all tests in parallel.
+        Returns a list of the results
+        """
+        result_futures = [
+            self.executor.submit(
+                self.run_test, f, print_failed=print_failed, print_format=print_format
+            )
+            for f in all_files
+        ]
+        results = []
+        for future in concurrent.futures.as_completed(result_futures):
+            dc_result = future.result()
+            results.append(dc_result)
+            print(dc_result.message)
+        return results
+
+    def run_serial(self, all_files, print_failed, print_format):
+        """
+        Runs all tests in serial.
+        Returns a list of the results
+        """
+        results = []
+        for f in all_files:
+            dc_result = self.run_test(
+                f, print_failed=print_failed, print_format=print_format
+            )
+            results.append(dc_result)
+            print(dc_result.message)
+        return results
+
+    def gen_expectation(self, sql_file: Path, force=False):
         """
         Executes the query for a data_check test
         and stores the result in the csv file.
         """
         expect_result = self.get_expect_file(sql_file)
-        if not expect_result.exists():
+        if not expect_result.exists() or force:
             result = self.run_query(sql_file.read_text(encoding="UTF-8"))
             result.to_csv(expect_result, index=False)
             print(f"expectation written to {expect_result}")
         else:
             print(f"expectation skipped for {expect_result}")
 
-    def generate_expectations(self, files: List[Path]):
+    def generate_expectations(self, files: List[Path], force=False):
         """
         Generated a expected results file for each file if it doesn't exists yet.
         """
         for sql_file in self.expand_files(files):
-            self.gen_expectation(sql_file)
+            self.gen_expectation(sql_file, force)
 
     def test_connection(self) -> bool:
         """
         Returns True if we can connect to the database.
         Mainly for integration tests.
         """
-        engine = create_engine(
-            self.connection, **{**self.get_db_params(), **{"pool_pre_ping": True}}
-        )
+        engine = self.get_engine(extra_params={"pool_pre_ping": True})
         try:
             engine.connect()
             return True
